==> slices_cli-0.1.37-py3-none-any.whl: slices_cli/experiment.py <==
"""Experiment CLI."""

from datetime import datetime
from typing import Annotated, Optional

import typer
from pytz import UTC
from rich.table import Table
from slices_expauth_client.exceptions import ApiException
from slices_expauth_client.models.experiment_create import ExperimentCreate
from slices_expauth_client.models.experiment_update import ExperimentUpdate
from slices_projauth_client.exceptions import NotFoundException as ProjectNotFoundException

from . import console, err_console
from .context import Context
from .util.pagination import get_next_page_marker
from .utils import datetime_with_local_timezone, friendly_datetime, get_error_detail, is_experiment_id, is_project_id

experiment_app = typer.Typer(short_help="Experiments", no_args_is_help=True)


@experiment_app.command(name="create")
def create_experiment(
    ctx: Context,
    name: Annotated[str, typer.Option("--name", "-n", help="Name of the experiment.")],
    expiration: Annotated[datetime, typer.Option("--expiration", "--exp", "-e")],
    description: Annotated[Optional[str], typer.Option("--description", "--desc", "-d")] = None,
    project_id: Annotated[Optional[str], typer.Option("--project-id", "-p")] = None,
):
    """Create a new experiment."""

    if project_id:
        if not is_project_id(project_id):
            project = ctx.obj.api.get_project_by_name(project_id)

            if not project:
                err_console.print(f"[red]Could not find project with name {project_id} :magnifying_glass_tilted_left:")
                raise typer.Exit(1)

            project_id = project.id

        project_token = ctx.obj.api.get_project_jwt(project_id)
    else:
        project_token = ctx.obj.get_current_project_token()

        if not project_token:
            err_console.print(
                "[bold red]No default project set. "
                "Please specify a default project with `slices project use <name>`, or pass a project id. :x:"
            )
            raise typer.Exit(1)

    try:
        experiment = ctx.obj.api.experiments_api.create_experiment(
            ExperimentCreate(
                friendly_name=name,
                description=description,
                expires_at=datetime_with_local_timezone(expiration),
                project_token=project_token,
            )
        )

    except ApiException as err:
        err_console.print("[bold red]Could not create experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f"[green]:sparkles: Created experiment [bold]{experiment.friendly_name}[/bold] "
        f"that expires on [bold]{friendly_datetime(experiment.expires_at)}[/]",
    )
    console.print(f"Experiment ID: {experiment.id}")


@experiment_app.command(name="list")
def list_experiments(  # noqa: C901
    ctx: Context,
    include_expired: Annotated[
        Optional[bool], typer.Option("--include-expired", "-e", help="Also show expired experiments.")
    ] = False,
    include_archived: Annotated[
        Optional[bool], typer.Option("--include-archived", "-A", help="Also show archived experiments.")
    ] = False,
    all: Annotated[Optional[bool], typer.Option("--all", "-a", help="Show all experiments.")] = None,
    all_projects: Annotated[
        Optional[bool], typer.Option("--all-projects", "-P", help="Show experiments from all your projects.")
    ] = None,
    show_ids: Annotated[bool, typer.Option("--show-ids", "-I", help="Show internal IDs.")] = False,
    name: Annotated[Optional[str], typer.Option("--name", "-n", help="Name of the experiment to list.")] = None,
):
    """List experiments."""

    if all is True:
        include_expired = True
        include_archived = True

    project_id = ctx.obj.get_current_project_id()
    if all_projects:
        project_id = None

    project_id_to_name_map = {}
    if project_id is None:
        projects = ctx.obj.api.projects_api.get_projects()
        project_id_to_name_map = {project.id: project.name for project in projects.projects}

    experiments = ctx.obj.api.experiments_api.get_experiments(
        include_expired=include_expired,
        include_archived=include_archived,
        project_id=project_id,
        friendly_name=name,
    )

    projects = ctx.obj.api.projects_api.get_projects()

    columns = ["Name", "Description", "Created at", "Expires at"]
    if include_archived:
        columns.append("Archived")

    if project_id is None:
        columns.insert(1, "Project")
    if show_ids:
        columns.insert(0, "ID")

    title = "Experiments"
    if project_id:
        project = ctx.obj.api.projects_api.get_project(project_id)
        title = f"Experiments in project [bold]{project.name}[/]"

    table = Table(*columns, title=title)

    for exp in experiments.experiments:
        exp_expires_at = friendly_datetime(exp.expires_at)
        if exp.expires_at < datetime.now(UTC):
            exp_expires_at = f"[red]{exp_expires_at}"

        row = [
            exp.friendly_name,
            exp.description,
            friendly_datetime(exp.created_at),
            exp_expires_at,
        ]
        if include_archived:
            row.append("Yes" if exp.archived else "No")
        if not project_id:
            row.insert(1, project_id_to_name_map.get(exp.project_id, exp.project_id))

        if show_ids:
            row.insert(0, exp.id)

        table.add_row(*row, style="red" if exp.archived else "")

    console.print(table)


def _get_experiment(ctx, experiment_id_or_name: str):
    if is_experiment_id(experiment_id_or_name):
        return ctx.obj.api.experiments_api.get_experiment(experiment_id=experiment_id_or_name)

    experiments = ctx.obj.api.experiments_api.get_experiments(
        project_id=ctx.obj.require_project_id(),
        friendly_name=experiment_id_or_name,
    )

    if not len(experiments.experiments) > 0:
        err_console.print(
            f"[red]Could not find experiment with name {experiment_id_or_name} :magnifying_glass_tilted_left:"
        )
        raise typer.Exit(1)

    return experiments.experiments[0]


@experiment_app.command("show")
def show_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment to show.")],
):
    """Show experiment info."""
    experiment = _get_experiment(ctx, name)

    project = None
    try:
        project = ctx.obj.api.projects_api.get_project(experiment.project_id)
    except ProjectNotFoundException:
        err_console.print("[orange]Could not fetch project info :warning:")

    table = Table(box=None)
    table.add_column("Name", justify="right")
    table.add_column("Value")
    table.add_row("ID", experiment.id)
    table.add_row("Name", experiment.friendly_name)
    table.add_row("Project ID", experiment.project_id)
    table.add_row("Project name", project.name if project else "** Unknown **")
    table.add_row("Created on", friendly_datetime(experiment.created_at))
    table.add_row("Expires on", friendly_datetime(experiment.expires_at))
    table.add_row("Deleted", "Yes" if experiment.archived else "No")

    console.print(table)


GENERIC_SITE_IDS = {"f4fwrapper": "be-gent1-bi-baremetal1", "vmlab": "be-gent1-bi-vm1"}


@experiment_app.command("list")
def list_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment to show.")],
    include_expired: Annotated[bool, typer.Option()] = True,
    include_terminated: Annotated[bool, typer.Option()] = True,
    show_ids: Annotated[bool, typer.Option("--show-ids", "-I", help="Show internal IDs.")] = False,
):
    """Show experiment resources."""

    experiment = _get_experiment(ctx, name)

    resources_page = ctx.obj.api.experiments_api.get_experiment_resources(
        experiment_id=experiment.id,
        include_expired=include_expired,
        include_terminated=include_terminated,
    )

    cols = [
        "Friendly Name",
        "Site",
        "Type",
        "Status",
        "Description",
        "Created At",
        "Expires At",
        "Terminated At",
    ]
    if show_ids:
        cols.insert(0, "ID")

    table = Table(
        *cols,
        title=f"Resources of experiment {experiment.friendly_name}",
    )

    while resources_page:
        for resource in resources_page.experiment_resources:
            row = [
                resource.friendly_name,
                GENERIC_SITE_IDS.get(resource.site_id, resource.site_id),
                resource.type,
                resource.status,
                resource.description,
                friendly_datetime(resource.created_at),
                friendly_datetime(resource.expires_at) if resource.expires_at else "",
                friendly_datetime(resource.terminated_at) if resource.terminated_at else "",
            ]
            if show_ids:
                row.insert(0, resource.id)

            table.add_row(*row)
        marker = get_next_page_marker(resources_page.experiment_resource_links)
        resources_page = (
            ctx.obj.api.experiments_api.get_experiment_resources(
                experiment_id=experiment.id,
                include_expired=include_expired,
                include_terminated=include_terminated,
                marker=marker,
            )
            if marker
            else None
        )

    console.print(table)


@experiment_app.command("renew")
def renew_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(help="Name or ID of experiment that must be renewed.")],
    expiration: Annotated[datetime, typer.Argument(help="New expiration date.")],
):
    """Extend the duration of an experiment."""
    if is_experiment_id(name):
        experiment_id = name
    else:
        experiments = ctx.obj.api.experiments_api.get_experiments(
            project_id=ctx.obj.require_project_id(),
            friendly_name=name,
        )

        if not len(experiments.experiments) > 0:
            err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
            raise typer.Exit(1)

        experiment_id = experiments.experiments[0].id

    try:
        renewed_experiment = ctx.obj.api.experiments_api.update_experiment(
            experiment_id, ExperimentUpdate(expires_at=datetime_with_local_timezone(expiration))
        )
    except ApiException as err:
        err_console.print("[bold red]Could not renew experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":recycle: Renewed experiment '{renewed_experiment.friendly_name}' "
        f"with ID '{renewed_experiment.id}' until {friendly_datetime(renewed_experiment.expires_at)}"
    )


@experiment_app.command("delete")
def delete_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment that must be deleted.")],
    force: Annotated[bool, typer.Option("--force", "-f", help="Do not ask for confirmation", is_flag=True)] = False,
):
    """Delete an experiment by name or ID."""
    if is_experiment_id(name):
        experiment_id = name
    else:
        experiments = ctx.obj.api.experiments_api.get_experiments(
            project_id=ctx.obj.require_project_id(),
            friendly_name=name,
        )

        if not len(experiments.experiments) > 0:
            err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
            raise typer.Exit(1)

        experiment_id = experiments.experiments[0].id

        if not force:
            typer.confirm(f"Are you sure you want to delete experiment {name} with ID '{experiment_id}'?", abort=True)

    try:
        deleted_experiment = ctx.obj.api.experiments_api.delete_experiment(experiment_id)
    except ApiException as err:
        err_console.print("[bold red]Could not delete experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":wastebasket: Deleted experiment '{deleted_experiment.friendly_name}' with ID '{deleted_experiment.id}'"
    )

==> slices_cli-0.1.39-py3-none-any.whl: slices_cli/experiment.py <==
"""Experiment CLI."""

from datetime import datetime
from typing import Annotated, Optional

import jwt
import typer
from pytz import UTC
from requests.exceptions import HTTPError
from rich.table import Table
from slices_expauth_client.exceptions import ApiException
from slices_expauth_client.models.experiment_create import ExperimentCreate
from slices_expauth_client.models.experiment_update import ExperimentUpdate
from slices_projauth_client.exceptions import NotFoundException as ProjectNotFoundException

from . import console, err_console
from .context import Context
from .util.pagination import get_next_page_marker
from .utils import datetime_with_local_timezone, friendly_datetime, get_error_detail, is_experiment_id, is_project_id

experiment_app = typer.Typer(short_help="Experiments", no_args_is_help=True)


@experiment_app.command(name="create")
def create_experiment(
    ctx: Context,
    name: Annotated[str, typer.Option("--name", "-n", help="Name of the experiment.")],
    expiration: Annotated[datetime, typer.Option("--expiration", "--exp", "-e")],
    description: Annotated[Optional[str], typer.Option("--description", "--desc", "-d")] = None,
    project_id: Annotated[Optional[str], typer.Option("--project-id", "-p")] = None,
):
    """Create a new experiment."""

    if project_id:
        if not is_project_id(project_id):
            project = ctx.obj.api.get_project_by_name(project_id)

            if not project:
                err_console.print(f"[red]Could not find project with name {project_id} :magnifying_glass_tilted_left:")
                raise typer.Exit(1)

            project_id = project.id

        project_token = ctx.obj.api.get_project_jwt(project_id)
    else:
        project_token = ctx.obj.get_current_project_token()

        if not project_token:
            err_console.print(
                "[bold red]No default project set. "
                "Please specify a default project with `slices project use <name>`, or pass a project id. :x:"
            )
            raise typer.Exit(1)

    try:
        experiment = ctx.obj.api.experiments_api.create_experiment(
            ExperimentCreate(
                friendly_name=name,
                description=description,
                expires_at=datetime_with_local_timezone(expiration),
                project_token=project_token,
            )
        )

    except ApiException as err:
        err_console.print("[bold red]Could not create experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f"[green]:sparkles: Created experiment [bold]{experiment.friendly_name}[/bold] "
        f"that expires on [bold]{friendly_datetime(experiment.expires_at)}[/]",
    )
    console.print(f"Experiment ID: {experiment.id}")


@experiment_app.command(name="list")
def list_experiments(  # noqa: C901
    ctx: Context,
    include_expired: Annotated[
        Optional[bool], typer.Option("--include-expired", "-e", help="Also show expired experiments.")
    ] = False,
    include_archived: Annotated[
        Optional[bool], typer.Option("--include-archived", "-A", help="Also show archived experiments.")
    ] = False,
    all: Annotated[Optional[bool], typer.Option("--all", "-a", help="Show all experiments.")] = None,
    all_projects: Annotated[
        Optional[bool], typer.Option("--all-projects", "-P", help="Show experiments from all your projects.")
    ] = None,
    show_ids: Annotated[bool, typer.Option("--show-ids", "-I", help="Show internal IDs.")] = False,
    name: Annotated[Optional[str], typer.Option("--name", "-n", help="Name of the experiment to list.")] = None,
):
    """List experiments."""

    if all is True:
        include_expired = True
        include_archived = True

    project_id = ctx.obj.get_current_project_id()
    if all_projects:
        project_id = None

    project_id_to_name_map = {}
    if project_id is None:
        projects = ctx.obj.api.projects_api.get_projects()
        project_id_to_name_map = {project.id: project.name for project in projects.projects}

    experiments = ctx.obj.api.experiments_api.get_experiments(
        include_expired=include_expired,
        include_archived=include_archived,
        project_id=project_id,
        friendly_name=name,
    )

    projects = ctx.obj.api.projects_api.get_projects()

    columns = ["Name", "Description", "Created at", "Expires at"]
    if include_archived:
        columns.append("Archived")

    if project_id is None:
        columns.insert(1, "Project")
    if show_ids:
        columns.insert(0, "ID")

    title = "Experiments"
    if project_id:
        project = ctx.obj.api.projects_api.get_project(project_id)
        title = f"Experiments in project [bold]{project.name}[/]"

    table = Table(*columns, title=title)

    for exp in experiments.experiments:
        exp_expires_at = friendly_datetime(exp.expires_at)
        if exp.expires_at < datetime.now(UTC):
            exp_expires_at = f"[red]{exp_expires_at}"

        row = [
            exp.friendly_name,
            exp.description,
            friendly_datetime(exp.created_at),
            exp_expires_at,
        ]
        if include_archived:
            row.append("Yes" if exp.archived else "No")
        if not project_id:
            row.insert(1, project_id_to_name_map.get(exp.project_id, exp.project_id))

        if show_ids:
            row.insert(0, exp.id)

        table.add_row(*row, style="red" if exp.archived else "")

    console.print(table)


def _get_experiment(ctx, experiment_id_or_name: str):
    if is_experiment_id(experiment_id_or_name):
        return ctx.obj.api.experiments_api.get_experiment(experiment_id=experiment_id_or_name)

    experiments = ctx.obj.api.experiments_api.get_experiments(
        project_id=ctx.obj.require_project_id(),
        friendly_name=experiment_id_or_name,
    )

    if not len(experiments.experiments) > 0:
        err_console.print(
            f"[red]Could not find experiment with name {experiment_id_or_name} :magnifying_glass_tilted_left:"
        )
        raise typer.Exit(1)

    return experiments.experiments[0]


@experiment_app.command("show")
def show_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment to show.")],
):
    """Show experiment info."""
    experiment = _get_experiment(ctx, name)

    project = None
    try:
        project = ctx.obj.api.projects_api.get_project(experiment.project_id)
    except ProjectNotFoundException:
        err_console.print("[orange]Could not fetch project info :warning:")

    table = Table(box=None)
    table.add_column("Name", justify="right")
    table.add_column("Value")
    table.add_row("ID", experiment.id)
    table.add_row("Name", experiment.friendly_name)
    table.add_row("Project ID", experiment.project_id)
    table.add_row("Project name", project.name if project else "** Unknown **")
    table.add_row("Created on", friendly_datetime(experiment.created_at))
    table.add_row("Expires on", friendly_datetime(experiment.expires_at))
    table.add_row("Deleted", "Yes" if experiment.archived else "No")

    console.print(table)


GENERIC_SITE_IDS = {"f4fwrapper": "be-gent1-bi-baremetal1", "vmlab": "be-gent1-bi-vm1"}


@experiment_app.command("list-resources")
def list_experiment_resources(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment to show.")],
    include_expired: Annotated[bool, typer.Option()] = True,
    include_terminated: Annotated[bool, typer.Option()] = True,
    show_ids: Annotated[bool, typer.Option("--show-ids", "-I", help="Show internal IDs.")] = False,
):
    """Show experiment resources."""

    experiment = _get_experiment(ctx, name)

    resources_page = ctx.obj.api.experiments_api.get_experiment_resources(
        experiment_id=experiment.id,
        include_expired=include_expired,
        include_terminated=include_terminated,
    )

    cols = [
        "Friendly Name",
        "Site",
        "Type",
        "Status",
        "Description",
        "Created At",
        "Expires At",
        "Terminated At",
    ]
    if show_ids:
        cols.insert(0, "ID")

    table = Table(
        *cols,
        title=f"Resources of experiment {experiment.friendly_name}",
    )

    while resources_page:
        for resource in resources_page.experiment_resources:
            row = [
                resource.friendly_name,
                GENERIC_SITE_IDS.get(resource.site_id, resource.site_id),
                resource.type,
                resource.status,
                resource.description,
                friendly_datetime(resource.created_at),
                friendly_datetime(resource.expires_at) if resource.expires_at else "",
                friendly_datetime(resource.terminated_at) if resource.terminated_at else "",
            ]
            if show_ids:
                row.insert(0, resource.id)

            table.add_row(*row)
        marker = get_next_page_marker(resources_page.experiment_resource_links)
        resources_page = (
            ctx.obj.api.experiments_api.get_experiment_resources(
                experiment_id=experiment.id,
                include_expired=include_expired,
                include_terminated=include_terminated,
                marker=marker,
            )
            if marker
            else None
        )

    console.print(table)


@experiment_app.command("renew")
def renew_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(help="Name or ID of experiment that must be renewed.")],
    expiration: Annotated[datetime, typer.Argument(help="New expiration date.")],
):
    """Extend the duration of an experiment."""
    if is_experiment_id(name):
        experiment_id = name
    else:
        experiments = ctx.obj.api.experiments_api.get_experiments(
            project_id=ctx.obj.require_project_id(),
            friendly_name=name,
        )

        if not len(experiments.experiments) > 0:
            err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
            raise typer.Exit(1)

        experiment_id = experiments.experiments[0].id

    try:
        renewed_experiment = ctx.obj.api.experiments_api.update_experiment(
            experiment_id, ExperimentUpdate(expires_at=datetime_with_local_timezone(expiration))
        )
    except ApiException as err:
        err_console.print("[bold red]Could not renew experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":recycle: Renewed experiment '{renewed_experiment.friendly_name}' "
        f"with ID '{renewed_experiment.id}' until {friendly_datetime(renewed_experiment.expires_at)}"
    )


@experiment_app.command("delete")
def delete_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment that must be deleted.")],
    force: Annotated[bool, typer.Option("--force", "-f", help="Do not ask for confirmation", is_flag=True)] = False,
):
    """Delete an experiment by name or ID."""
    if is_experiment_id(name):
        experiment_id = name
    else:
        experiments = ctx.obj.api.experiments_api.get_experiments(
            project_id=ctx.obj.require_project_id(),
            friendly_name=name,
        )

        if not len(experiments.experiments) > 0:
            err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
            raise typer.Exit(1)

        experiment_id = experiments.experiments[0].id

        if not force:
            typer.confirm(f"Are you sure you want to delete experiment {name} with ID '{experiment_id}'?", abort=True)

    try:
        deleted_experiment = ctx.obj.api.experiments_api.delete_experiment(experiment_id)
    except ApiException as err:
        err_console.print("[bold red]Could not delete experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":wastebasket: Deleted experiment '{deleted_experiment.friendly_name}' with ID '{deleted_experiment.id}'"
    )


@experiment_app.command("jwt", help="Get credentials for a specific experiment.", hidden=True)
def retrieve_jwt(
    ctx: Context,
    name: str,
    decode: Annotated[bool, typer.Option("--decode", "-d", help="Decode the token.")] = False,
):
    """Print experiment credentials."""
    experiment = _get_experiment(ctx, name)

    try:
        token = ctx.obj.api.get_experiment_jwt(experiment.id)

        if decode:
            console.print(jwt.decode(token, options={"verify_signature": False}))
        else:
            console.print(token)
    except HTTPError as e:
        try:
            error_response = e.response.json()
            print(f"[bold red]Error: Could not get token. :x:[/]\n[red]{error_response['error_description']}")
        except:  # noqa: E722
            print("[bold red]Error: Could not get token. :x:")
            print(f"[red]{str(e.response.text)}")
        raise typer.Exit(1) from None
